/*
 * Listener.hpp
 *
 *  Created on: 22.02.2012
 *      Author: heber
 */

#ifndef LISTENER_HPP_
#define LISTENER_HPP_

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <boost/asio.hpp>

#include "Connection.hpp"

/** Class Listener implements a class listening on a given port for incoming
 * connections.
 *
 * In order to use, inherit it and implement the virtual function handle_Accept().
 * \code
 *   class foo : public Listener {
 *     ...
 *     void handle_Accept(const boost::system::error_code& e, connection_ptr conn)
 *     {
 *       if (!e) {
 *         conn->async_write(job,
 *          boost::bind(&<functionpointer>,
 *          boost::asio::placeholders::error, conn));
 *      }
 *       ...
 *     }
 *   };
 * \endcode
 * where async_write or async_read expect some bound function taking again
 * the error and the connection as arguments (i.e. a cascade of handlers).
 *
 */
class Listener
{
public:
  Listener(boost::asio::io_service& io_service, unsigned short port);
  ~Listener();

  void initiateSocket();
  void closeSocket();

  enum Exitflag_t{
    OkFlag = 0,
    ErrorFlag = 255
  };

  /** Getter for Exitflag.
   *
   * @return Exitflag of operations
   */
  size_t getExitflag() const
  {
    return Exitflag;
  }

protected:
  /// Handle completion for initially acceptig connection
  virtual void handle_Accept(const boost::system::error_code& e, connection_ptr conn)=0;

  // Exit flag on program exit
  enum Exitflag_t Exitflag;

private:
  /// The acceptor object used to accept incoming worker socket connections.
  boost::asio::ip::tcp::acceptor acceptor_;

  //!> Connection object
  connection_ptr connection;
};

#endif /* LISTENER_HPP_ */
